@@ -31,7 +31,7 @@ from hypothesis.internal.compat import ceil, hbytes, int_from_bytes
 from hypothesis.internal.conjecture.data import (
     ConjectureData,
     ConjectureResult,
-    GenerationParameters,
+    DataObserver,
     Overrun,
     Status,
     StopTest,
@@ -42,7 +42,7 @@ from hypothesis.internal.conjecture.datatree import (
     TreeRecordingObserver,
 )
 from hypothesis.internal.conjecture.junkdrawer import clamp
-from hypothesis.internal.conjecture.pareto import NO_SCORE, ParetoFront
+from hypothesis.internal.conjecture.pareto import NO_SCORE, ParetoFront, ParetoOptimiser
 from hypothesis.internal.conjecture.shrinker import Shrinker, sort_key
 from hypothesis.internal.healthcheck import fail_health_check
 from hypothesis.reporting import base_report
@@ -553,11 +553,6 @@ class ConjectureRunner(object):
             if (
                 self.valid_examples >= self.settings.max_examples
                 or self.call_count >= max(self.settings.max_examples * 10, 1000)
-                or (
-                    self.best_examples_of_observed_targets
-                    and self.valid_examples * 2 >= self.settings.max_examples
-                    and self.should_optimise
-                )
             ):  # pragma: no cover
                 return False
 
@@ -578,24 +573,6 @@ class ConjectureRunner(object):
                 self.first_bug_found_at + 1000, self.last_bug_found_at * 2
             )
 
-        # GenerationParameters are a set of decisions we make that are global
-        # to the whole test case, used to bias the data generation in various
-        # ways. This is an approach very very loosely inspired by the paper
-        # "Swarm testing." by Groce et al. in that it induces deliberate
-        # correlation between otherwise independent decisions made during the
-        # generation process.
-        #
-        # More importantly the generation is designed to make certain scenarios
-        # more likely (e.g. small examples, duplicated values), which can help
-        # or hurt in terms of finding interesting things. Whenever the result
-        # of our generation is a bad test case, for whatever definition of
-        # "bad" we like (currently, invalid or too large), we ditch the
-        # parameter early. This allows us to potentially generate good test
-        # cases significantly more often than we otherwise would, by selecting
-        # for parameters that make them more likely.
-        parameter = GenerationParameters(self.random)
-        count = 0
-
         # We attempt to use the size of the minimal generated test case starting
         # from a given novel prefix as a guideline to generate smaller test
         # cases for an initial period, by restriscting ourselves to test cases
@@ -612,30 +589,32 @@ class ConjectureRunner(object):
         # it has failed too many times in a row.
         consecutive_zero_extend_is_invalid = 0
 
+        # We control growth during initial example generation, for two
+        # reasons:
+        #
+        # * It gives us an opportunity to find small examples early, which
+        #   gives us a fast path for easy to find bugs.
+        # * It avoids low probability events where we might end up
+        #   generating very large examples during health checks, which
+        #   on slower machines can trigger HealthCheck.too_slow.
+        #
+        # The heuristic we use is that we attempt to estimate the smallest
+        # extension of this prefix, and limit the size to no more than
+        # an order of magnitude larger than that. If we fail to estimate
+        # the size accurately, we skip over this prefix and try again.
+        #
+        # We need to tune the example size based on the initial prefix,
+        # because any fixed size might be too small, and any size based
+        # on the strategy in general can fall afoul of strategies that
+        # have very different sizes for different prefixes.
+        small_example_cap = clamp(10, self.settings.max_examples // 10, 50)
+
+        optimise_at = max(self.settings.max_examples // 2, small_example_cap + 1)
+        ran_optimisations = False
+
         while should_generate_more():
             prefix = self.generate_novel_prefix()
             assert len(prefix) <= BUFFER_SIZE
-
-            # We control growth during initial example generation, for two
-            # reasons:
-            #
-            # * It gives us an opportunity to find small examples early, which
-            #   gives us a fast path for easy to find bugs.
-            # * It avoids low probability events where we might end up
-            #   generating very large examples during health checks, which
-            #   on slower machines can trigger HealthCheck.too_slow.
-            #
-            # The heuristic we use is that we attempt to estimate the smallest
-            # extension of this prefix, and limit the size to no more than
-            # an order of magnitude larger than that. If we fail to estimate
-            # the size accurately, we skip over this prefix and try again.
-            #
-            # We need to tune the example size based on the initial prefix,
-            # because any fixed size might be too small, and any size based
-            # on the strategy in general can fall afoul of strategies that
-            # have very different sizes for different prefixes.
-            small_example_cap = clamp(10, self.settings.max_examples // 10, 50)
-
             if (
                 self.valid_examples <= small_example_cap
                 and self.call_count <= 5 * small_example_cap
@@ -666,7 +645,7 @@ class ConjectureRunner(object):
                 # starting from the new novel prefix that has discovered.
                 try:
                     trial_data = self.new_conjecture_data(
-                        prefix=prefix, parameter=parameter, max_length=max_length
+                        prefix=prefix, max_length=max_length
                     )
                     self.tree.simulate_test_function(trial_data)
                     continue
@@ -687,20 +666,110 @@ class ConjectureRunner(object):
             else:
                 max_length = BUFFER_SIZE
 
-            data = self.new_conjecture_data(
-                prefix=prefix, parameter=parameter, max_length=max_length
-            )
+            data = self.new_conjecture_data(prefix=prefix, max_length=max_length)
 
             self.test_function(data)
 
-            count += 1
+            # A thing that is often useful but rarely happens by accident is
+            # to generate the same value at multiple different points in the
+            # test case.
+            #
+            # Rather than make this the responsibility of individual strategies
+            # we implement a small mutator that just takes parts of the test
+            # case with the same label and tries replacing one of them with a
+            # copy of the other and tries running it. If we've made a good
+            # guess about what to put where, this will run a similar generated
+            # test case with more duplication.
             if (
-                data.status < Status.VALID
-                or len(data.buffer) * 2 >= BUFFER_SIZE
-                or count > 5
+                # An OVERRUN doesn't have enough information about the test
+                # case to mutate, so we just skip those.
+                data.status >= Status.INVALID
+                # This has a tendency to trigger some weird edge cases during
+                # generation so we don't let it run until we're done with the
+                # health checks.
+                and self.health_check_state is None
             ):
-                count = 0
-                parameter = GenerationParameters(self.random)
+                initial_calls = self.call_count
+                failed_mutations = 0
+                while (
+                    should_generate_more()
+                    # We implement fairly conservative checks for how long we
+                    # we should run mutation for, as it's generally not obvious
+                    # how helpful it is for any given test case.
+                    and self.call_count <= initial_calls + 5
+                    and failed_mutations <= 5
+                ):
+                    groups = defaultdict(list)
+                    for ex in data.examples:
+                        groups[ex.label, ex.depth].append(ex)
+
+                    groups = [v for v in groups.values() if len(v) > 1]
+
+                    if not groups:
+                        break
+
+                    group = self.random.choice(groups)
+
+                    ex1, ex2 = sorted(
+                        self.random.sample(group, 2), key=lambda i: i.index
+                    )
+                    assert ex1.end <= ex2.start
+
+                    replacements = [data.buffer[e.start : e.end] for e in [ex1, ex2]]
+
+                    replacement = self.random.choice(replacements)
+
+                    try:
+                        # We attempt to replace both the the examples with
+                        # whichever choice we made. Note that this might end
+                        # up messing up and getting the example boundaries
+                        # wrong - labels matching are only a best guess as to
+                        # whether the two are equivalent - but it doesn't
+                        # really matter. It may not achieve the desired result
+                        # but it's still a perfectly acceptable choice sequence.
+                        # to try.
+                        new_data = self.cached_test_function(
+                            data.buffer[: ex1.start]
+                            + replacement
+                            + data.buffer[ex1.end : ex2.start]
+                            + replacement
+                            + data.buffer[ex2.end :],
+                            # We set error_on_discard so that we don't end up
+                            # entering parts of the tree we consider redundant
+                            # and not worth exploring.
+                            error_on_discard=True,
+                            extend=BUFFER_SIZE,
+                        )
+                    except ContainsDiscard:
+                        failed_mutations += 1
+                        continue
+
+                    if (
+                        new_data.status >= data.status
+                        and data.buffer != new_data.buffer
+                        and all(
+                            k in new_data.target_observations
+                            and new_data.target_observations[k] >= v
+                            for k, v in data.target_observations.items()
+                        )
+                    ):
+                        data = new_data
+                        failed_mutations = 0
+                    else:
+                        failed_mutations += 1
+
+            # Although the optimisations are logically a distinct phase, we
+            # actually normally run them as part of example generation. The
+            # reason for this is that we cannot guarantee that optimisation
+            # actually exhausts our budget: It might finish running and we
+            # discover that actually we still could run a bunch more test cases
+            # if we want.
+            if (
+                self.valid_examples >= max(small_example_cap, optimise_at)
+                and not ran_optimisations
+            ):
+                ran_optimisations = True
+                self.optimise_targets()
 
     def optimise_targets(self):
         """If any target observations have been made, attempt to optimise them
@@ -711,24 +780,37 @@ class ConjectureRunner(object):
             prev_calls = self.call_count
             for target, data in list(self.best_examples_of_observed_targets.items()):
                 self.new_optimiser(data, target).run()
-            # This shows up as a missing branch for some reason but is definitely
-            # covered if you add an assertion to check.
-            if self.interesting_examples or (prev_calls == self.call_count):
+
+            if self.interesting_examples:
                 break
 
+            self.pareto_optimise()
+
+            if prev_calls == self.call_count:
+                break
+
+    def pareto_optimise(self):
+        if self.pareto_front is not None:
+            ParetoOptimiser(self).run()
+
     def _run(self):
         self.reuse_existing_examples()
         self.generate_new_examples()
-        self.optimise_targets()
+
+        # We normally run the targeting phase mixed in with the generate phase,
+        # but if we've been asked to run it but not generation then we have to
+        # run it explciitly on its own here.
+        if Phase.generate not in self.settings.phases:
+            self.optimise_targets()
         self.shrink_interesting_examples()
         self.exit_with(ExitReason.finished)
 
-    def new_conjecture_data(self, prefix, parameter, max_length=BUFFER_SIZE):
+    def new_conjecture_data(self, prefix, max_length=BUFFER_SIZE, observer=None):
         return ConjectureData(
             prefix=prefix,
-            parameter=parameter,
             max_length=max_length,
-            observer=self.tree.new_observer(),
+            random=self.random,
+            observer=observer or self.tree.new_observer(),
         )
 
     def new_conjecture_data_for_buffer(self, buffer):
@@ -822,15 +904,24 @@ class ConjectureRunner(object):
 
         return Optimiser(self, example, target)
 
-    def cached_test_function(self, buffer):
+    def cached_test_function(self, buffer, error_on_discard=False, extend=0):
         """Checks the tree to see if we've tested this buffer, and returns the
         previous result if we have.
 
         Otherwise we call through to ``test_function``, and return a
         fresh result.
+
+        If ``error_on_discard`` is set to True this will raise ``ContainsDiscard``
+        in preference to running the actual test function. This is to allow us
+        to skip test cases we expect to be redundant in some cases. Note that
+        it may be the case that we don't raise ``ContainsDiscard`` even if the
+        result has discards if we cannot determine from previous runs whether
+        it will have a discard.
         """
         buffer = hbytes(buffer)[:BUFFER_SIZE]
 
+        max_length = min(BUFFER_SIZE, len(buffer) + extend)
+
         def check_result(result):
             assert result is Overrun or (
                 isinstance(result, ConjectureResult) and result.status != Status.OVERRUN
@@ -842,16 +933,34 @@ class ConjectureRunner(object):
         except KeyError:
             pass
 
-        rewritten, status = self.tree.rewrite(buffer)
+        if error_on_discard:
+
+            class DiscardObserver(DataObserver):
+                def kill_branch(self):
+                    raise ContainsDiscard()
+
+            observer = DiscardObserver()
+        else:
+            observer = DataObserver()
+
+        dummy_data = self.new_conjecture_data(
+            prefix=buffer, max_length=max_length, observer=observer
+        )
 
         try:
-            result = check_result(self.__data_cache[rewritten])
-        except KeyError:
+            self.tree.simulate_test_function(dummy_data)
+        except PreviouslyUnseenBehaviour:
             pass
         else:
-            assert result.status != Status.OVERRUN or result is Overrun
-            self.__data_cache[buffer] = result
-            return result
+            if dummy_data.status > Status.OVERRUN:
+                dummy_data.freeze()
+                try:
+                    return self.__data_cache[dummy_data.buffer]
+                except KeyError:
+                    pass
+            else:
+                self.__data_cache[buffer] = Overrun
+                return Overrun
 
         # We didn't find a match in the tree, so we need to run the test
         # function normally. Note that test_function will automatically
@@ -859,17 +968,11 @@ class ConjectureRunner(object):
 
         result = None
 
-        if status != Status.OVERRUN:
-            data = self.new_conjecture_data_for_buffer(buffer)
-            self.test_function(data)
-            result = check_result(data.as_result())
-            assert status is None or result.status == status, (status, result.status)
-            status = result.status
-        if status == Status.OVERRUN:
-            result = Overrun
-
-        assert result is not None
-
+        data = self.new_conjecture_data(
+            prefix=max((buffer, dummy_data.buffer), key=len), max_length=max_length,
+        )
+        self.test_function(data)
+        result = check_result(data.as_result())
         self.__data_cache[buffer] = result
         return result
 
@@ -883,3 +986,7 @@ class ConjectureRunner(object):
         result = str(event)
         self.events_to_strings[event] = result
         return result
+
+
+class ContainsDiscard(Exception):
+    pass
